import json
from urllib.parse import urlencode

import pytest
from django.contrib.auth import get_user_model
from django.http import HttpResponseBadRequest, HttpResponseForbidden
from django.test import Client
from django.urls import reverse
from testapp.models import Ausgabe

from mizdb_tomselect.views import FILTERBY_VAR, PAGE_SIZE, PAGE_VAR, SEARCH_LOOKUP_VAR, SEARCH_VAR


@pytest.fixture
def obj():
    return Ausgabe.objects.create(name="Test", num="1", lnum="2", jahr="3")


@pytest.fixture(autouse=True)
def not_found():
    # Add a 'control' object that should never be included in the search results.
    return Ausgabe.objects.create(name="Not Found")


@pytest.fixture
def pages():
    # Create enough data for multiple pages.
    return [
        Ausgabe.objects.create(name=f"2022-{i + 1:02}", num=i + 1, lnum=100 + i, jahr="2022")
        for i in range(PAGE_SIZE * 2)
    ]


@pytest.fixture
def noperms_user():
    return get_user_model().objects.create_user(username="noperms", password="bar")


@pytest.mark.django_db
class TestAutocompleteView:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.url = reverse("autocomplete")
        self.model_label = f"{Ausgabe._meta.app_label}.{Ausgabe._meta.model_name}"

    def test_get_contains_result(self, admin_client, obj):
        """The response of a search query should contain the expected result."""
        query_string = urlencode({"model": self.model_label, SEARCH_VAR: "test", SEARCH_LOOKUP_VAR: "name__icontains"})
        response = admin_client.get(f"{self.url}?{query_string}")
        data = json.loads(response.content)
        assert data["results"] == list(Ausgabe.objects.filter(pk=obj.pk).values())

    @pytest.mark.parametrize("page_number,has_more", [(1, True), (2, False)])
    def test_context_pagination(self, admin_client, pages, page_number, has_more):
        """
        The response of a search query should contain context items for
        pagination.
        """
        request_data = {
            "model": self.model_label,
            SEARCH_VAR: "2022",
            PAGE_VAR: str(page_number),
            SEARCH_LOOKUP_VAR: "name__icontains",
        }
        response = admin_client.get(self.url, data=request_data)
        data = json.loads(response.content)
        assert data["page"] == page_number
        assert data["has_more"] == has_more

    def test_post_creates_new_object(self, admin_client):
        """A successful POST request should create a new model object."""
        request_data = {"model": self.model_label, "name": "New Ausgabe", "create-field": "name"}
        response = admin_client.post(self.url, data=request_data)
        assert response.status_code == 200
        assert Ausgabe.objects.filter(name="New Ausgabe").exists()

    def test_post_context_contains_object_data(self, admin_client):
        """
        The context of a successful POST request should contain data about the
        created item.
        """
        request_data = {
            "model": self.model_label,
            "name": "New Ausgabe",
            "create-field": "name",
        }
        response = admin_client.post(self.url, data=request_data)
        new = Ausgabe.objects.get(name="New Ausgabe")
        data = json.loads(response.content)
        assert data["pk"] == new.pk
        assert data["text"] == str(new)

    def test_post_no_permission(self, client, noperms_user):
        """
        Users that do not have 'add' permission should have the POST request
        denied.
        """
        client.force_login(noperms_user)
        response = client.post(self.url, data={"model": self.model_label})
        assert isinstance(response, HttpResponseForbidden)

    def test_post_user_not_authenticated(self, client):
        """POST requests by unauthenticated users should be denied."""
        client.logout()
        response = client.post(self.url, data={"model": self.model_label})
        assert isinstance(response, HttpResponseForbidden)

    def test_post_create_field_data_missing(self, admin_client):
        """
        POST requests that do not provide a value for the 'create field' should
        be denied.
        """
        request_data = {
            "model": self.model_label,
            # 'name' is missing
            "create-field": "name",
        }
        response = admin_client.post(self.url, data=request_data)
        assert isinstance(response, HttpResponseBadRequest)

    @pytest.mark.parametrize("has_csrf_token", (True, False))
    def test_post_csrf(self, admin_user, has_csrf_token):
        """Assert that requests without a CSRF token are denied."""
        client = Client(enforce_csrf_checks=True)
        client.force_login(admin_user)
        client.get(reverse("csrf"))  # have the csrf middleware set the cookie
        token = client.cookies["csrftoken"]

        request_data = {
            "model": self.model_label,
            "name": "New Ausgabe",
            "create-field": "name",
            "csrfmiddlewaretoken": token.coded_value if has_csrf_token else "",
        }
        response = client.post(self.url, data=request_data)
        if not has_csrf_token:
            assert response.status_code == 403
        else:
            assert response.status_code == 200

    def test_get_no_search_term(self, client):
        """Assert that a GET request without a search term still returns results."""
        query_string = urlencode({"model": self.model_label, SEARCH_LOOKUP_VAR: "name__icontains"})
        response = client.get(f"{self.url}?{query_string}")
        data = json.loads(response.content)
        assert data["results"]

    def test_get_with_filter_by(self, client, obj):
        """
        Assert that a GET request with a filterBy value returns the expected
        results.
        """
        query_string = urlencode(
            {"model": self.model_label, SEARCH_LOOKUP_VAR: "name__icontains", FILTERBY_VAR: "lnum=2"}
        )
        response = client.get(f"{self.url}?{query_string}")
        data = json.loads(response.content)
        assert len(data["results"]) == 1
        assert data["results"][0]["id"] == obj.pk

    def test_get_no_filter_by(self, client):
        """
        Assert that a GET request returns no results when a required filterBy
        has no value.
        """
        query_string = urlencode(
            {"model": self.model_label, SEARCH_LOOKUP_VAR: "name__icontains", FILTERBY_VAR: "lnum="}
        )
        response = client.get(f"{self.url}?{query_string}")
        data = json.loads(response.content)
        assert not data["results"]



import pytest
from testapp.models import Ausgabe

from mizdb_tomselect.widgets import MIZSelect, MIZSelectTabular


class WidgetTestCase:
    widget_class = None

    @pytest.fixture(autouse=True)
    def make_widget(self):
        def _make_widget(model=Ausgabe, **kwargs):
            return self.widget_class(model, **kwargs)

        self.make_widget = _make_widget


class TestMIZSelect(WidgetTestCase):
    widget_class = MIZSelect

    def test_optgroups_no_initial_choices(self):
        """Assert that the widget is rendered without any options."""
        context = self.make_widget().get_context("ausgabe", None, {})
        assert not context["widget"]["optgroups"]

    def test_build_attrs(self):
        """Assert that the required HTML attributes are added."""
        widget = self.make_widget(
            model=Ausgabe,
            url="dummy_url",
            value_field="pk",
            label_field="num",
            create_field="the_create_field",
            multiple=True,
            changelist_url="changelist_page",
            add_url="add_page",
        )
        attrs = widget.build_attrs({})
        assert attrs["is-tomselect"]
        assert attrs["is-multiple"]
        assert attrs["data-autocomplete-url"] == "/dummy/url/"
        assert attrs["data-model"] == f"{Ausgabe._meta.app_label}.{Ausgabe._meta.model_name}"
        assert attrs["data-value-field"] == "pk"
        assert attrs["data-label-field"] == "num"
        assert attrs["data-create-field"] == "the_create_field"
        assert attrs["data-changelist-url"] == "/testapp/changelist/"
        assert attrs["data-add-url"] == "/testapp/add/"

    @pytest.mark.parametrize(
        "static_file",
        ("mizselect.css", "tom-select.bootstrap5.css", "mizselect.js"),
    )
    def test_media(self, static_file):
        """Assert that the necessary static files are included."""
        assert static_file in str(self.make_widget().media)


class TestTabularMIZSelect(WidgetTestCase):
    widget_class = MIZSelectTabular

    def test_build_attrs(self):
        """Assert that the required HTML attributes are added."""
        widget = self.make_widget(
            model=Ausgabe,
            extra_columns={"jahr": "Jahr", "num": "Nummer", "lnum": "lfd.Nummer"},
            value_field_label="Primary Key",
            label_field_label="Ausgabe",
        )
        attrs = widget.build_attrs({})
        assert attrs["is-tabular"]
        assert attrs["data-value-field-label"] == "Primary Key"
        assert attrs["data-label-field-label"] == "Ausgabe"
        assert attrs["data-extra-headers"] == '["Jahr", "Nummer", "lfd.Nummer"]'
        assert attrs["data-extra-columns"] == '["jahr", "num", "lnum"]'



import json

from django import forms
from django.urls import reverse


class MIZSelect(forms.Select):
    """
    A TomSelect widget with model object choices.

    The TomSelect element will be configured using custom data attributes on
    the select element, which are provided by the widget's `build_attrs` method.
    """

    def __init__(
        self,
        model,
        url="autocomplete",
        search_lookup="name__icontains",
        value_field="id",
        label_field="name",
        create_field="",
        multiple=False,
        changelist_url="",
        add_url="",
        filter_by=(),
        **kwargs,
    ):
        """
        Instantiate a MIZSelect widget.

        Args:
            model: the django model that the choices are derived from
            url: the URL pattern name of the view that serves the choices and
              handles requests from the TomSelect element
            search_lookup: a Django field lookup to use with the given search
              term to filter the results
            value_field: the name of the model field that corresponds to the
              choice value of an option (f.ex. 'id')
            label_field: the name of the model field that corresponds to the
              human-readable value of an option (f.ex. 'name')
            create_field: the name of the model field used to create new
              model objects with
            multiple: if True, allow selecting multiple options
            changelist_url: URL name of the changelist view for this model
            add_url: URL name of the add view for this model
            filter_by: a 2-tuple (form_field_name, field_lookup) to filter the
              results against the value of the form field using the given
              Django field lookup. For example:
               ('foo', 'bar__id') => results.filter(bar__id=data['foo'])
            kwargs: additional keyword arguments passed to forms.Select
        """
        self.model = model
        self.url = url
        self.search_lookup = search_lookup
        self.create_field = create_field
        self.value_field = value_field
        self.label_field = label_field
        self.multiple = multiple
        self.changelist_url = changelist_url
        self.add_url = add_url
        self.filter_by = filter_by
        super().__init__(**kwargs)

    def optgroups(self, name, value, attrs=None):
        return []  # Never provide any options; let the view serve the options.

    def get_url(self):
        """Hook to specify the autocomplete URL."""
        return reverse(self.url)

    def get_add_url(self):
        """Hook to specify the URL to the model's add page."""
        if self.add_url:
            return reverse(self.add_url)

    def get_changelist_url(self):
        """Hook to specify the URL the model's changelist."""
        if self.changelist_url:
            return reverse(self.changelist_url)

    def build_attrs(self, base_attrs, extra_attrs=None):
        """Build HTML attributes for the widget."""
        attrs = super().build_attrs(base_attrs, extra_attrs)
        opts = self.model._meta
        attrs.update(
            {
                "is-tomselect": True,
                "is-multiple": self.multiple,
                "data-autocomplete-url": self.get_url(),
                "data-model": f"{opts.app_label}.{opts.model_name}",
                "data-search-lookup": self.search_lookup,
                "data-value-field": self.value_field,
                "data-label-field": self.label_field,
                "data-create-field": self.create_field,
                "data-changelist-url": self.get_changelist_url() or "",
                "data-add-url": self.get_add_url() or "",
                "data-filter-by": json.dumps(list(self.filter_by)),
            }
        )
        return attrs

    class Media:
        css = {
            "all": ["vendor/tom-select/css/tom-select.bootstrap5.css", "mizdb_tomselect/css/mizselect.css"],
        }
        js = ["mizdb_tomselect/js/mizselect.js"]


class MIZSelectTabular(MIZSelect):
    """A MIZSelect widget that displays results in a table with a table header."""

    def __init__(self, *args, extra_columns=None, value_field_label="ID", label_field_label="Object", **kwargs):
        """
        Instantiate a MIZSelectTabular widget.

        Args:
            extra_columns: a mapping of <model field names> to <column labels>
              for additional columns. The field name tells TomSelect what
              values to look up on a model object result for a given column.
              The label is the table header label for a given column.
            value_field_label: table header label for the value field column
            label_field_label: table header label for the label field column
            args: additional positional arguments passed to MIZSelect
            kwargs: additional keyword arguments passed to MIZSelect
        """
        super().__init__(*args, **kwargs)
        self.value_field_label = value_field_label
        self.label_field_label = label_field_label
        self.extra_columns = extra_columns or {}

    def build_attrs(self, base_attrs, extra_attrs=None):
        """Build HTML attributes for the widget."""
        attrs = super().build_attrs(base_attrs, extra_attrs)
        attrs.update(
            {
                "is-tabular": True,
                "data-value-field-label": self.value_field_label,
                "data-label-field-label": self.label_field_label,
                "data-extra-headers": json.dumps(list(self.extra_columns.values())),
                "data-extra-columns": json.dumps(list(self.extra_columns.keys())),
            }
        )
        return attrs



import pytest
from django.contrib.auth import get_user_model
from django.urls import reverse
from playwright.sync_api import expect
from testapp.models import Ausgabe, Magazin

from mizdb_tomselect.views import PAGE_SIZE

# Mark all tests in this module as end-to-end tests (excluded from running by default):
pytestmark = pytest.mark.e2e


@pytest.fixture
def magazin():
    return Magazin.objects.create(name="Testmagazin")


@pytest.fixture(autouse=True)
def other_magazin():
    return Magazin.objects.create(name="Other")


@pytest.fixture(autouse=True)  # TODO: is autouse=True necessary?
def data(magazin):
    return [
        Ausgabe.objects.create(name=f"2022-{i + 1:02}", num=i + 1, lnum=100 + i, jahr="2022", magazin=magazin)
        for i in range(PAGE_SIZE * 3)
    ]


@pytest.fixture
def first_object(data):
    return data[0]


@pytest.fixture(autouse=True)
def noperms_user():
    return get_user_model().objects.create(username="noperms", password="sadface")


@pytest.fixture
def user(admin_user, username):
    """
    Return the user with the given username.

    If username is empty or 'admin', return a pytest-django admin_user.
    """
    if not username or username == "admin":
        return admin_user
    return get_user_model().objects.get(username=username)


@pytest.fixture
def login(client, user):
    """Log in the given user."""
    client.force_login(user)
    return client


@pytest.fixture
def logged_in(login, context):
    """
    Log in a user and add the session cookie for the logged-in user to the
    current context.
    """
    auth_cookie = login.cookies["sessionid"]
    pw_cookie = {
        "name": auth_cookie.key,
        "value": auth_cookie.value,
        "path": auth_cookie["path"],
        "domain": auth_cookie["domain"] or "localhost",
    }
    context.add_cookies([pw_cookie])


# NOTE: must not override original page fixture or tests can't be run against
# multiple browsers
@pytest.fixture
def _page(page, live_server, view_name):
    page.goto(live_server.url + reverse(view_name))
    return page


@pytest.fixture
def ts_wrapper(_page):
    wrapper = _page.locator(".ts-wrapper")
    wrapper.wait_for()
    return wrapper


@pytest.fixture
def wrapper_focus(_page, ts_wrapper):
    with _page.expect_event("requestfinished"):
        ts_wrapper.click()


@pytest.fixture
def search_input(_page, wrapper_focus):
    search_input = _page.locator(".dropdown-input")
    search_input.wait_for()
    return search_input


@pytest.fixture
def search(_page, search_input):
    with _page.expect_event("requestfinished"):
        search_input.fill("2022")


def get_dropdown_items(_page):
    """Return all elements in the dropdown."""
    return _page.locator(".ts-dropdown-content > *")


def get_select_options(_page):
    """Return all selectable options."""
    return _page.locator("[data-selectable][role=option]")


def get_last_dropdown_item(_page):
    item = None
    for item in get_dropdown_items(_page).all():
        # Have to wait for each option to be attached, otherwise
        # test_virtual_scroll fails for the 'create' view.
        item.wait_for(state="attached")
    return item


@pytest.mark.parametrize("view_name", ["simple"])
def test_initially_empty(_page, view_name):
    """Assert that the list of options is empty initially."""
    expect(get_dropdown_items(_page)).to_have_count(0)


@pytest.mark.django_db
@pytest.mark.parametrize("view_name", ["simple"])
def test_load_first_results_on_focus(_page, view_name, wrapper_focus):
    """Assert that the first _page of options is loaded when the select gets focus."""
    expect(get_select_options(_page)).to_have_count(PAGE_SIZE)


@pytest.mark.django_db
@pytest.mark.parametrize("view_name", ["simple", "multiple", "tabular", "create"])
def test_virtual_scroll(_page, view_name, search, data):
    """
    Assert that the user can scroll to the bottom of the options to load more
    options from the backend until there are no more search results to load.
    """
    expect(get_select_options(_page)).to_have_count(PAGE_SIZE)

    # Second _page:
    with _page.expect_event("requestfinished"):
        get_last_dropdown_item(_page).scroll_into_view_if_needed()
    expect(get_select_options(_page)).to_have_count(PAGE_SIZE * 2)

    # Last _page:
    with _page.expect_event("requestfinished"):
        get_last_dropdown_item(_page).scroll_into_view_if_needed()
    expect(get_select_options(_page)).to_have_count(len(data))
    expect(get_last_dropdown_item(_page)).to_have_text("Keine weiteren Ergebnisse")


@pytest.mark.django_db
@pytest.mark.parametrize("view_name", ["multiple"])
class TestSelectMultiple:
    @pytest.fixture
    def select_two(self, _page, search):
        """Select the first two options."""
        options = get_select_options(_page).all()
        options[0].click()
        options[1].click()

    @pytest.fixture
    def ts_control(self, _page):
        return _page.locator(".ts-control")

    @pytest.fixture
    def selected(self, _page, select_two, ts_control):
        """Return the locator for the selected options."""
        return ts_control.locator(".item")

    def test_select_multiple(self, _page, selected):
        """Assert that the user can select multiple options."""
        expect(selected).to_have_count(2)

    def test_has_remove_button(self, _page, selected):
        """Assert that each selected option has a remove button."""
        for option in selected.all():
            expect(option.locator(".remove")).to_be_attached()

    def test_has_clear_button(self, _page, ts_control):
        """Assert that the ts control has a clear all button."""
        expect(ts_control.locator(".clear-button")).to_have_count(1)


@pytest.mark.django_db
@pytest.mark.parametrize("view_name", ["tabular"])
class TestTabularSelect:
    @pytest.fixture
    def dropdown_header(self, _page, wrapper_focus):
        """Return the dropdown header."""
        return _page.locator(".ts-dropdown .dropdown-header")

    @pytest.fixture
    def header_columns(self, _page, dropdown_header):
        """Return the column divs of the dropdown header."""
        return dropdown_header.locator(".row > *")

    @pytest.fixture
    def option_columns(self, _page, wrapper_focus):
        """Return the column divs of a select option."""
        return get_select_options(_page).first.locator("*")

    def test_has_dropdown_header(self, _page, dropdown_header):
        """Assert that the dropdown has the expected table header."""
        expect(dropdown_header).to_be_attached()

    def test_header_columns(self, _page, header_columns):
        """Assert that the dropdown header has the expected columns."""
        expect(header_columns).to_have_count(5)
        id_col, label_col, jahr_col, num_col, lnum_col = header_columns.all()
        expect(id_col).to_have_class("col-1")
        expect(id_col).to_have_text("ID")
        expect(label_col).to_have_class("col-5")
        expect(label_col).to_have_text("Ausgabe")
        expect(jahr_col).to_have_class("col")
        expect(jahr_col).to_have_text("Jahr")
        expect(num_col).to_have_class("col")
        expect(num_col).to_have_text("Nummer")
        expect(lnum_col).to_have_class("col")
        expect(lnum_col).to_have_text("lfd.Nummer")

    def test_options_have_columns(self, _page, option_columns, first_object):
        """Assert that the data of the options are assigned to columns."""
        expect(option_columns).to_have_count(5)
        id_col, label_col, jahr_col, num_col, lnum_col = option_columns.all()
        expect(id_col).to_have_class("col-1")
        expect(id_col).to_have_text(str(first_object.pk))
        expect(label_col).to_have_class("col-5")
        expect(label_col).to_have_text(first_object.name)
        expect(jahr_col).to_have_class("col")
        expect(jahr_col).to_have_text(str(first_object.jahr))
        expect(num_col).to_have_class("col")
        expect(num_col).to_have_text(str(first_object.num))
        expect(lnum_col).to_have_class("col")
        expect(lnum_col).to_have_text(str(first_object.lnum))


################################################################################
# Dropdown Footer
################################################################################


@pytest.fixture
def dropdown_footer(_page, wrapper_focus):
    """Return the dropdown footer."""
    return _page.locator(".dropdown-footer")


@pytest.fixture
def add_button(_page, dropdown_footer):
    """Return the add button in the dropdown footer."""
    return dropdown_footer.locator(".add-btn")


@pytest.fixture
def changelist_button(dropdown_footer):
    """Return the changelist button in the dropdown footer."""
    return dropdown_footer.locator(".cl-btn")


@pytest.mark.parametrize("view_name,has_footer", [("add", True), ("changelist", True), ("simple", False)])
def test_has_footer(view_name, has_footer, dropdown_footer, context):
    """
    Assert that a footer div is added to the dropdown content if the select
    element declares a 'changelist' URL or a 'add' URL.
    """
    if has_footer:
        # Note that the footer will be attached for an 'add' URL even if the
        # user has no 'add' permission.
        expect(dropdown_footer).to_be_attached()
    else:
        expect(dropdown_footer).not_to_be_attached()


@pytest.mark.parametrize("view_name", ["add"])
@pytest.mark.parametrize("username", ["noperms"])
def test_add_button_invisible_when_no_permission(
    logged_in,
    add_button,
    view_name,
    username,
):
    """
    Assert that the add button is invisible if the user does not have
    'add' permission.
    """
    expect(add_button).not_to_be_visible()


@pytest.mark.django_db
@pytest.mark.parametrize("view_name", ["add"])
@pytest.mark.parametrize("username", ["admin"])
class TestFooterAddButton:
    def test_has_visible_add_button(self, logged_in, add_button):
        """Assert that the dropdown footer contains a visible 'add' button."""
        expect(add_button).to_be_visible()

    def test_add_button_text_changes_on_typing(self, logged_in, _page, add_button, search_input):
        """
        Assert that the text of the add button updates along with the user
        typing in a search term.
        """
        expect(add_button).to_have_text("Hinzufügen")
        with _page.expect_event("requestfinished"):
            search_input.fill("202")
        expect(add_button).to_have_text("'202' hinzufügen...")
        with _page.expect_event("requestfinished"):
            search_input.fill("2022")
        expect(add_button).to_have_text("'2022' hinzufügen...")

    def test_add_button_click_no_search_term(self, logged_in, _page, add_button, live_server):
        """
        Assert that clicking the add button with no search term given opens the
        'add' _page in a new tab.
        """

        def requests_add_page(request):
            """Return whether the request is for the add _page."""
            return request.url == live_server.url + reverse("add_page")

        with _page.expect_popup(requests_add_page):
            add_button.click()

    def test_add_button_click_with_search_term(self, logged_in, _page, add_button, search_input, live_server):
        """
        Assert that clicking the add button with a search term given starts a
        POST request to create a new object instead of opening the 'add' _page.
        """
        with _page.expect_request_finished():
            search_input.fill("2022-99")
        with _page.expect_response(live_server.url + reverse("ac"), timeout=1000) as response_info:
            add_button.click()
        response = response_info.value
        data = response.json()
        assert data["text"] == "2022-99"
        assert data["pk"]

    def test_add_button_successful_creation(self, logged_in, _page, add_button, search_input):
        """
        If the POST request to create a new object was successful, the created
        item should be immediately selected.
        """
        with _page.expect_request_finished():
            search_input.fill("2022-99")
        with _page.expect_request_finished():
            add_button.click()
        expect(_page.locator(".ts-control .item")).to_have_text("2022-99")
        expect(_page.locator(".dropdown-content")).not_to_be_visible()


@pytest.mark.django_db
@pytest.mark.parametrize("view_name", ["changelist"])
class TestFooterChangelistButton:
    def test_has_visible_changelist_button(self, changelist_button):
        """Assert that the dropdown footer contains a visible 'changelist' button."""
        expect(changelist_button).to_be_visible()

    def test_changelist_query_string_contains_search_term(self, _page, changelist_button, search_input):
        """
        Assert that the URL to the changelist contains the current search term
        in the query string.
        """
        assert changelist_button.get_attribute("href") == reverse("changelist_page")
        with _page.expect_request_finished():
            search_input.fill("2022")
        assert "q=2022" in changelist_button.get_attribute("href")


@pytest.fixture
def magazin_select(_page):
    select = _page.get_by_label("Magazin")
    select.wait_for()
    return select


@pytest.mark.django_db
@pytest.mark.parametrize("view_name", ["filtered"])
def test_options_filtered(_page, magazin_select, magazin, other_magazin, ts_wrapper):
    """
    Assert that the options of the 'ausgabe' field are filtered with the values
    of the forwarded field 'magazin'.
    """
    magazin_select.select_option(value=str(other_magazin.pk))
    with _page.expect_request_finished():
        ts_wrapper.click()
    expect(get_select_options(_page)).to_have_count(0)
    magazin_select.focus()
    magazin_select.select_option(value=str(magazin.pk))
    with _page.expect_request_finished():
        ts_wrapper.click()
    expect(get_select_options(_page)).not_to_have_count(0)



from django import http, views
from django.apps import apps
from django.contrib.auth import get_permission_codename

SEARCH_VAR = "q"
SEARCH_LOOKUP_VAR = "sl"
FILTERBY_VAR = "f"
PAGE_VAR = "p"
PAGE_SIZE = 20


class AutocompleteView(views.generic.list.BaseListView):
    """Base list view for queries from TomSelect select elements."""

    paginate_by = PAGE_SIZE
    page_kwarg = PAGE_VAR

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        request_data = getattr(request, request.method)
        self.model = apps.get_model(request_data["model"])
        self.create_field = request_data.get("create-field")
        self.search_lookup = request.GET.get(SEARCH_LOOKUP_VAR)

    def apply_filter_by(self, request, queryset):
        """
        Filter the result queryset with values set by other form fields.

        If the widget was set up with a `filter_by` parameter, the request will
        include the `FILTERBY_VAR` parameter, indicating that the results must
        be filtered against the lookup and value provided by `FILTERBY_VAR`.
        If `FILTERBY_VAR` is present, but no value is set, return an empty
        queryset.
        """
        if FILTERBY_VAR not in request.GET:
            return queryset
        else:
            lookup, value = request.GET[FILTERBY_VAR].split("=")
            if not value:
                # A filter was set up for this autocomplete, but no filter value
                # was provided; return an empty queryset.
                return queryset.none()
            return queryset.filter(**{lookup: value})

    def search(self, request, queryset, q):
        """Filter the result queryset against the search term."""
        return queryset.filter(**{self.search_lookup: q})

    def order_queryset(self, queryset):
        """Order the result queryset."""
        ordering = self.model._meta.ordering or ["id"]
        return queryset.order_by(*ordering)

    def get_results(self, request):
        """
        Search for objects that match the search parameters and return the
        results as a values() queryset.
        """
        queryset = self.get_queryset()
        q = request.GET.get(SEARCH_VAR, "")
        if q or FILTERBY_VAR in request.GET:
            queryset = self.apply_filter_by(request, queryset)
            queryset = self.search(request, queryset, q)
        return self.order_queryset(queryset.values())

    def get(self, request, *args, **kwargs):
        queryset = self.get_results(request)
        page_size = self.get_paginate_by(queryset)
        _, page, queryset, is_paginated = self.paginate_queryset(queryset, page_size)
        data = {
            "results": list(page.object_list),
            "page": page.number,
            "has_more": page.has_next(),
            "show_create_option": self.has_add_permission(request),
        }
        return http.JsonResponse(data)

    def has_add_permission(self, request):
        """Return True if the user has the permission to add a model object."""
        if not request.user.is_authenticated:
            return False

        opts = self.model._meta
        codename = get_permission_codename("add", opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def create_object(self, data):
        """Create a new object with the given data."""
        return self.model.objects.create(**{self.create_field: data[self.create_field]})

    def post(self, request, *args, **kwargs):
        if not self.has_add_permission(request):
            return http.HttpResponseForbidden()
        if request.POST.get(self.create_field) is None:
            return http.HttpResponseBadRequest()
        obj = self.create_object(request.POST)
        return http.JsonResponse({"pk": obj.pk, "text": str(obj)})
